import logging

import click

from . import options
from .. import core, track, update, utils
from ..trio_utils import coro
from .base import CatchAllExceptionsCommand

logger = logging.getLogger(__name__)
console = utils.getConsole()


@click.command(
    name="update",
    help="Generate EPUB files for new parts of all tracked series (or specific "
    "series if a URL argument is passed)",
    cls=CatchAllExceptionsCommand,
)
@click.argument("jnc_url", metavar="(JNOVEL_CLUB_URL?)", required=False)
@options.login_option
@options.password_option
@options.output_option
@options.byvolume_option
@options.images_option
@options.raw_content_option
@options.no_replace_chars_option
@options.css_option
@click.option(
    "-s",
    "--sync",
    "is_sync",
    is_flag=True,
    help=(
        "Flag to sync tracked series based on series followed on J-Novel Club and "
        "update the new ones from the beginning of the series"
    ),
)
@click.option(
    "-w",
    "--whole",
    "is_whole_volume",
    is_flag=True,
    envvar="JNCEP_WHOLE",
    help=(
        "Flag to indicate whether the whole volume should be regenerated when a "
        "new part is detected during the update"
    ),
)
@coro
async def update_tracked(
    jnc_url,
    email,
    password,
    output_dirpath,
    is_by_volume,
    is_extract_images,
    is_extract_content,
    is_not_replace_chars,
    style_css_path,
    is_sync,
    is_whole_volume,
):
    epub_generation_options = core.EpubGenerationOptions(
        output_dirpath,
        is_by_volume,
        is_extract_images,
        is_extract_content,
        is_not_replace_chars,
        style_css_path,
    )

    async with core.JNCEPSession(email, password) as session:
        track_manager = track.TrackConfigManager()
        tracked_series = track_manager.read_tracked_series()
        if len(tracked_series) == 0:
            console.warning(
                "There are no tracked series! Use the 'jncep track add' command "
                "first."
            )
            return

        new_synced = None
        if is_sync:
            console.status("Fetch followed series from J-Novel Club...")
            follows = await session.api.fetch_follows()
            new_synced, _ = await track.sync_series_forward(
                session, follows, tracked_series, False
            )

        if jnc_url:
            console.status(f"Update '{jnc_url}'...")

            is_tracking_updated = await update.update_url_series(
                session,
                jnc_url,
                epub_generation_options,
                tracked_series,
                is_sync,
                new_synced,
                is_whole_volume,
            )

        else:
            console.status("Update all series...")

            is_tracking_updated = await update.update_all_series(
                session,
                epub_generation_options,
                tracked_series,
                is_sync,
                new_synced,
                is_whole_volume,
            )

        if is_tracking_updated:
            track_manager.write_tracked_series(tracked_series)
            logger.debug("Data for tracked series sucessfully updated!")



import logging

import click

from . import options
from .. import core, jncweb, spec, utils
from ..trio_utils import coro
from .base import CatchAllExceptionsCommand

logger = logging.getLogger(__name__)
console = utils.getConsole()


@click.command(
    name="epub",
    help="Generate EPUB files for J-Novel Club pre-pub novels",
    cls=CatchAllExceptionsCommand,
)
@click.argument("jnc_url", metavar="JNOVEL_CLUB_URL", required=True)
@options.login_option
@options.password_option
@options.output_option
@click.option(
    "-s",
    "--parts",
    "part_spec",
    help=(
        "Specification of a range of parts to download in the form of "
        "<vol>[.part]:<vol>[.part] [default: All the content linked by "
        "the JNOVEL_CLUB_URL argument, either a single part, a whole volume "
        "or the whole series]"
    ),
)
@options.byvolume_option
@options.images_option
@options.raw_content_option
@options.no_replace_chars_option
@options.css_option
@coro
async def generate_epub(
    jnc_url,
    email,
    password,
    part_spec,
    output_dirpath,
    is_by_volume,
    is_extract_images,
    is_extract_content,
    is_not_replace_chars,
    style_css_path,
):
    # created by group
    epub_generation_options = core.EpubGenerationOptions(
        output_dirpath,
        is_by_volume,
        is_extract_images,
        is_extract_content,
        is_not_replace_chars,
        style_css_path,
    )

    async with core.JNCEPSession(email, password) as session:
        jnc_resource = jncweb.resource_from_url(jnc_url)

        if part_spec:
            console.info(f"Use part specification '[highlight]{part_spec}[/]'")
            part_spec_analyzed = spec.analyze_part_specs(part_spec)
        else:
            part_spec_analyzed = await core.to_part_spec(session, jnc_resource)

        console.status("Get content...")

        series = await core.resolve_series(session, jnc_resource)
        series_meta = await core.fill_meta(
            session, series, volume_callback=part_spec_analyzed.has_volume
        )

        # TODO event ? Think better logging system / notification to the user
        # to support maybe GUI

        has_unavailable_parts = False

        def part_filter(part):
            if part_spec_analyzed.has_part(part):
                if core.is_part_available(session.now, part):
                    return True
                else:
                    nonlocal has_unavailable_parts
                    has_unavailable_parts = True
            return False

        (
            volumes_to_download,
            parts_to_download,
        ) = core.relevant_volumes_and_parts_for_content(series_meta, part_filter)

        if not parts_to_download:
            console.error(
                "None of the requested parts are available! No EPUB will be generated.",
            )
            return

        if has_unavailable_parts:
            console.warning(
                "Some of the requested parts are not available for reading!",
            )

        volumes_for_cover = core.relevant_volumes_for_cover(
            volumes_to_download, epub_generation_options.is_by_volume
        )

        await core.fill_covers_and_content(
            session, volumes_for_cover, parts_to_download
        )

        console.status("Create EPUB...")

        await core.create_epub(
            series_meta,
            volumes_to_download,
            parts_to_download,
            epub_generation_options,
        )



import logging
from typing import List

import click
import dateutil.parser

from . import options
from .. import core, jncweb, track, utils
from ..trio_utils import coro
from ..utils import tryint
from .base import CatchAllExceptionsCommand

logger = logging.getLogger(__name__)
console = utils.getConsole()


@click.group(name="track", help="Track updates to a series")
def track_series():
    pass


@track_series.command(
    name="add", help="Add a new series for tracking", cls=CatchAllExceptionsCommand
)
@click.argument("jnc_url", metavar="JNOVEL_CLUB_URL", required=True)
@options.login_option
@options.password_option
@coro
async def add_track_series(jnc_url, email, password):
    async with core.JNCEPSession(email, password) as session:
        # TODO async read
        track_manager = track.TrackConfigManager()
        tracked_series = track_manager.read_tracked_series()

        console.status("Check tracking status...")

        jnc_resource = jncweb.resource_from_url(jnc_url)
        series = await core.resolve_series(session, jnc_resource)

        series_url = jncweb.url_from_series_slug(series.raw_data.slug)
        if series_url in tracked_series:
            console.warning(
                f"The series '[highlight]{series.raw_data.title}[/]' is "
                "already tracked!"
            )
            return

        await track.track_series(session, tracked_series, series)

        # TOO async write
        track_manager.write_tracked_series(tracked_series)


@track_series.command(
    name="sync",
    help="Sync list of series to track based on series followed on J-Novel Club "
    "website",
    cls=CatchAllExceptionsCommand,
)
@options.login_option
@options.password_option
@click.option(
    "-r",
    "--reverse",
    "is_reverse",
    is_flag=True,
    help=(
        "Flag to sync followed series on JNC website based on the series tracked with "
        "jncep"
    ),
)
@click.option(
    "-d",
    "--delete",
    "is_delete",
    is_flag=True,
    help="Flag to delete series not found on the sync source",
)
@coro
async def sync_series(email, password, is_reverse, is_delete):
    track_manager = track.TrackConfigManager()
    tracked_series = track_manager.read_tracked_series()

    async with core.JNCEPSession(email, password) as session:
        console.status("Fetch followed series from J-Novel Club...")
        follows: List[jncweb.JNCResource] = await session.api.fetch_follows()

        if is_reverse:
            console.status("Sync to J-Novel Club...")

            new_synced, del_synced = await track.sync_series_backward(
                session, follows, tracked_series, is_delete
            )

            if new_synced or del_synced:
                console.info(
                    "The list of followed series has been sucessfully updated!",
                    style="success",
                )
            else:
                console.info(
                    "Everything is already synced!",
                    style="success",
                )

        else:
            console.status("Sync tracked series from J-Novel Club...")

            new_synced, del_synced = await track.sync_series_forward(
                session, follows, tracked_series, is_delete
            )

            track_manager.write_tracked_series(tracked_series)

            if new_synced or del_synced:
                console.info(
                    "The list of tracked series has been sucessfully updated!",
                    style="success",
                )
            else:
                console.info(
                    "Everything is already synced!",
                    style="success",
                )


@track_series.command(
    name="rm", help="Remove a series from tracking", cls=CatchAllExceptionsCommand
)
@click.argument("jnc_url_or_index", metavar="JNOVEL_CLUB_URL_OR_INDEX", required=True)
@options.login_option
@options.password_option
@coro
async def rm_track_series(jnc_url_or_index, email, password):
    track_manager = track.TrackConfigManager()
    tracked_series = track_manager.read_tracked_series()

    index = tryint(jnc_url_or_index)
    if index is not None:
        index0 = index - 1
        if index0 < 0 or index0 >= len(tracked_series):
            console.warning(f"Index '{index}' is not valid! (Use 'track list')")
            return
        series_url_list = list(tracked_series.keys())
        series_url = series_url_list[index0]
    else:
        async with core.JNCEPSession(email, password) as session:
            console.status("Check tracking status...")
            jnc_resource = jncweb.resource_from_url(jnc_url_or_index)
            series = await core.resolve_series(session, jnc_resource)
            series_url = jncweb.url_from_series_slug(series.raw_data.slug)

            if series_url not in tracked_series:
                console.warning(
                    f"The series '[highlight]{series.raw_data.title}[/]' is not "
                    "tracked! (Use 'track list --details')"
                )
                return

    series_name = tracked_series[series_url].name

    del tracked_series[series_url]

    track_manager.write_tracked_series(tracked_series)

    console.info(
        f"The series '[highlight]{series_name}[/]' is no longer tracked",
        style="success",
    )


@track_series.command(
    name="list", help="List tracked series", cls=CatchAllExceptionsCommand
)
@click.option(
    "-t",
    "--details",
    "is_detail",
    is_flag=True,
    help="Flag to list the details of the tracked series (URL, date of last release)",
)
def list_track_series(is_detail):
    # TODO async ? zero utility
    track_manager = track.TrackConfigManager()
    tracked_series = track_manager.read_tracked_series()

    if len(tracked_series) > 0:
        console.info(f"{len(tracked_series)} series are tracked:")
        for index, (ser_url, ser_details) in enumerate(tracked_series.items()):
            details = None
            if ser_details.part_date:
                part_date = dateutil.parser.parse(ser_details.part_date)
                part_date_formatted = part_date.strftime("%b %d, %Y")
                details = f"{ser_details.part} [{part_date_formatted}]"
            elif ser_details.part == 0:
                details = "No part released"
            else:
                details = f"{ser_details.part}"

            msg = f"[[yellow]{index + 1}[/]] [green]{ser_details.name}[/]"
            if is_detail:
                msg += f" {ser_url} [red]{details}[/]"

            console.info(msg)
    else:
        console.warning("No series is tracked.")



import logging
import sys
import traceback

import click

from .. import utils

console = utils.getConsole()

logger = logging.getLogger(__package__)


class CatchAllExceptionsCommand(click.Command):
    def invoke(self, ctx):
        try:
            return super().invoke(ctx)
        except Exception as ex:
            raise UnrecoverableJNCEPError(str(ex), sys.exc_info())


class UnrecoverableJNCEPError(click.ClickException):
    def __init__(self, message, exc_info):
        super().__init__(message)
        self.exc_info = exc_info

    def show(self):
        emoji = ""
        if console.is_advanced():
            emoji = "\u274C "
        console.error(f"*** {emoji}An unrecoverable error occured ***")
        console.error(self.message)

        logger.debug(" ".join(traceback.format_exception(*self.exc_info)))



import os

import click

login_option = click.option(
    "-l",
    "--email",
    required=True,
    envvar="JNCEP_EMAIL",
    help="Login email for J-Novel Club account",
)

password_option = click.option(
    "-p",
    "--password",
    required=True,
    envvar="JNCEP_PASSWORD",
    help="Login password for J-Novel Club account",
)

output_option = click.option(
    "-o",
    "--output",
    "output_dirpath",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, writable=True),
    default=os.getcwd(),
    envvar="JNCEP_OUTPUT",
    help="Existing folder to write the output [default: The current directory]",
)

byvolume_option = click.option(
    "-v",
    "--byvolume",
    "is_by_volume",
    is_flag=True,
    envvar="JNCEP_BYVOLUME",
    help=(
        "Flag to indicate that the parts of different volumes shoud be output in "
        "separate EPUBs"
    ),
)

images_option = click.option(
    "-i",
    "--images",
    "is_extract_images",
    is_flag=True,
    envvar="JNCEP_IMAGES",
    help=(
        "Flag to indicate that the images of the novel should be extracted into "
        "the output folder"
    ),
)

raw_content_option = click.option(
    "-c",
    "--content",
    "is_extract_content",
    is_flag=True,
    envvar="JNCEP_CONTENT",
    help=(
        "Flag to indicate that the raw content of the parts should be extracted into "
        "the output folder"
    ),
)

# TODO rename alias to --noreplace (like --byvolume) or the opposite
# provide backward compatibility
no_replace_chars_option = click.option(
    "-n",
    "--no-replace",
    "is_not_replace_chars",
    is_flag=True,
    envvar="JNCEP_NOREPLACE",
    help=(
        "Flag to indicate that some unicode characters unlikely to be in an EPUB "
        "reader font should NOT be replaced and instead kept as is"
    ),
)


css_option = click.option(
    "-t",
    "--css",
    "style_css_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
    envvar="JNCEP_CSS",
    help="Path to custom CSS file for the EPUBs [default: The CSS provided by JNCEP]",
)
